#lang scheme

(require "find.ss"
         "test.ss")

(provide play)

;; find-new : hash-of-room hash-of-dest -> dest-or-#f
;;  Visited hash mapes visited to #f (not #t!)
;;  Look for a move that takes us to a new room
(define (find-new visited dests)
  (for/or ([pos (in-hash-keys dests)])
    (and (symbol? (car pos))
         (hash-ref visited (car pos) pos))))
(test (find-new #hash() #hash()) #f)
(test (find-new #hash() #hash(((1 1) . #t))) #f)
(test (find-new #hash() #hash(((1 1) . #t) ((U S) . #t))) '(U S))
(test (find-new #hash((U . #f)) #hash(((1 1) . #t) ((U S) . #t))) #f)

(define (should-roll? visited state me random?)
  (define (none? v) (equal? v #hash()))
  (let-values ([(roll-destss  nonroll-dests)
                (find-all-possible state me)])
    (cond
      ;; No way to move if we don't roll? Then roll.
      [(none? nonroll-dests) #t]
      ;; No way t move it we roll? Then don't roll.
      [(andmap none? roll-destss) #f]
      ;; Randomly pick, unless we're sensible
      [random? (zero? (random 2))]
      ;; More sensibly, don't roll if you can move
      ;;  directly to a new room (via a secret passage):
      [(find-new visited nonroll-dests) #f]
      ;; Otherwise, just roll:
      [else #t])))

(define (move visited state me roll random?)
  (let ([dests (find-dests state me roll)])
    (if (zero? (hash-count dests))
        ;; We can't move
        (list-ref state me)
        (or
         ;; Look for a move that takes us to a new room:
         (and (not random?) (find-new visited dests))
         ;; Otherwise, pick randomly:
         (list-ref
          (for/list ([pos (in-hash-keys dests)]) pos)
          (random (hash-count dests)))))))

(define (play host port name random?)
  (let-values ([(r w) (tcp-connect host port)])
    ;; Register our name:
    (fprintf w "~a\n" name)
    (flush-output w)
    ;; Loop, keeping track of rooms that we've visited:
    (let loop ([visited #hash()])
      ;; Wait until our turn:
      (let ([state (read r)]
            [me (read r)])
        (fprintf w (if (should-roll? visited state me random?)                       
                       "roll\n"
                       "no-roll\n"))
        (flush-output w)
        (let* ([roll (read r)]
               [dest (move visited state me roll random?)])
          (fprintf w "~a\n" dest)
          (flush-output w)
          ;; If we entered a room, keep track. Iterate.
          (loop (if (symbol? (car dest))
                    (hash-set visited (car dest) #f)
                    visited)))))))
