#lang scheme

(require "board.ss"
         "find.ss"
         "validate.ss")

(provide serve)

;; A player is
;;  (make-player input-port output-port string hash)
(define-struct player (r w number name visited))

(define (get-players l)
  (for/list ([i (in-range 6)]
             [name '("Scarlet" "Mustard" "White" "Green" "Peacock" "Plum")])
    (let-values ([(r w) (tcp-accept l)])
      (make-player r w 
                   i
                   (string-append (read-line r) " " name)
                   #hash()))))

;; check-must-roll : state nat -> string-or-#f
;;  Returns a required roll command, if any
(define (check-must-roll state me)
  (define (none? v) (equal? v #hash()))
  (let-values ([(roll-destss  nonroll-dests)
                (find-all-possible state me)])
    (cond
     [(none? nonroll-dests)
      (if (andmap none? roll-destss) 
          #f
          "roll")]
     [(andmap none? roll-destss) "no-roll"]
     [else #f])))

(define (serve port msg! state!)
  (define l (tcp-listen port 5 #t))
  (lambda ()
    (let ([state '((17 24) (24 17) (15 0) (10 0) (0 6) (0 19))])
      ;; Loop until someone wins. After we let the first player
      ;; move, shift the record to the end of the list.
      (let loop ([state state]
                 [players (get-players l)])
        (let ([p (car players)])
          ((with-handlers ([exn? (lambda (e)
                                   (error (format
                                           "during player ~s: ~a"
                                           (player-name p)
                                           (exn-message e))))])
             ;; Update display:
             (state! (format "Turn: ~a\n" (player-name p))
                     state)
             ;; Tell player to move:
             (fprintf (player-w p) "~s ~s\n" state (player-number p))
             (flush-output (player-w p))
             ;; Get roll request from player:
             (let* ([must-str (check-must-roll state (player-number p))]
                    [roll 
                     (let ([cmd (read-line (player-r p))])
                       (cond
                        [(and must-str (not (equal? must-str cmd)))
                         (error 'roll "bad roll choice: ~s" cmd)]
                        [(equal? cmd "roll")
                         (+ (add1 (random 6)) (add1 (random 6)))]
                        [(equal? cmd "no-roll")
                         0]
                        [else (error 'roll "bad roll replay: ~s" cmd)]))])
               ;; Report back roll:
               (fprintf (player-w p) "~s\n" roll)
               (flush-output (player-w p))
               ;; Get back player's destination: 
               (let* ([cmd (read-line (player-r p))]
                      [dest (read (open-input-string cmd))])
                 ;; Check the destination format:
                 (unless (and (list? dest)
                              (= 2 (length dest))
                              (or (and (number? (car dest))
                                       (number? (cadr dest)))
                                  (and (room? (car dest))
                                       (memq (cadr dest) '(S V)))))
                   (error 'move "bad move reply: ~s" cmd))
                 (msg! (format "~a ~a; moved to ~a"
                               (player-name p)
                               (if (zero? roll)
                                   "didn't roll"
                                   (format "rolled ~a" roll))
                               dest))
                 (let ([new-state (append (take state (player-number p))
                                          (list dest)
                                          (drop state (add1 (player-number p))))])
                   (unless (valid-move? state (player-number p) roll new-state)
                     (error 'move "illegal move: ~s" cmd))
                   (let ([p 
                          ;; Player moved into a new room?
                          (if (symbol? (car dest))
                              (struct-copy player
                                           p 
                                           [visited (hash-set (player-visited p)
                                                              (car dest) 
                                                              #t)])
                              p)])
                     ;; Check for winning condition to decide how to continue:
                     (if (= (hash-count (player-visited p)) 9)
                         (lambda () 
                           ;; game over
                           (state! (format "~a wins!" (player-name p))
                                   new-state))
                         (lambda ()
                           ;; next player
                           (loop new-state
                                 (append (cdr players) (list p))))))))))))))))
